@@ -4,17 +4,17 @@
 --
 
 {-# LANGUAGE MultiWayIf #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
 
-module System.FSNotify.OSX
-       ( FileListener(..)
-       , NativeManager
-       ) where
+module System.FSNotify.OSX (
+  FileListener(..)
+  , NativeManager
+  ) where
 
-import Control.Concurrent.Chan
-import Control.Concurrent.MVar
+import Control.Concurrent
 import Control.Monad
 import Data.Bits
-import Data.IORef (atomicModifyIORef, readIORef)
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Time.Clock (UTCTime, getCurrentTime)
@@ -29,7 +29,7 @@ import System.FilePath
 import qualified System.OSX.FSEvents as FSE
 
 
-data WatchData = WatchData FSE.EventStream EventChannel
+data WatchData = WatchData FSE.EventStream EventCallback
 
 type WatchMap = Map Unique WatchData
 data OSXManager = OSXManager (MVar WatchMap)
@@ -71,6 +71,7 @@ fsnEvents timestamp e = do
   -- putStrLn $ show ["Event", show e, "isDirectory", show isDirectory, "isFile", show isFile, "isModified", show isModified, "isCreated", show isCreated, "path", path e, "exists", show exists]
 
   return $ if | exists && isModified -> [Modified (path e) timestamp isDirectory]
+              | exists && isModifiedAttributes -> [ModifiedAttributes (path e) timestamp isDirectory]
               | exists && isCreated -> [Added (path e) timestamp isDirectory]
               | (not exists) && hasFlag e FSE.eventFlagItemRemoved -> [Removed (path e) timestamp isDirectory]
 
@@ -80,60 +81,51 @@ fsnEvents timestamp e = do
 
               | otherwise -> []
   where
-    isDirectory = hasFlag e FSE.eventFlagItemIsDir
+    isDirectory = if hasFlag e FSE.eventFlagItemIsDir then IsDirectory else IsFile
     isFile = hasFlag e FSE.eventFlagItemIsFile
     isCreated = hasFlag e FSE.eventFlagItemCreated
     isRenamed = hasFlag e FSE.eventFlagItemRenamed
-    isModified = hasFlag e FSE.eventFlagItemModified || hasFlag e FSE.eventFlagItemInodeMetaMod
+    isModified = hasFlag e FSE.eventFlagItemModified
+    isModifiedAttributes = hasFlag e FSE.eventFlagItemInodeMetaMod
     path = canonicalEventPath
     hasFlag event flag = FSE.eventFlags event .&. flag /= 0
 
-handleEvent :: Bool -> ActionPredicate -> EventChannel -> FilePath -> DebouncePayload -> FSE.Event -> IO ()
-handleEvent isRecursive actPred chan dirPath dbp fseEvent = do
+handleFSEEvent :: Bool -> ActionPredicate -> EventCallback -> FilePath -> FSE.Event -> IO ()
+handleFSEEvent isRecursive actPred callback dirPath fseEvent = do
   currentTime <- getCurrentTime
   events <- fsnEvents currentTime fseEvent
-  handleEvents isRecursive actPred chan dirPath dbp events
+  forM_ events $ \event ->
+    when (actPred event && (isRecursive || (isDirectlyInside dirPath event))) $
+      callback event
 
 -- | For non-recursive monitoring, test if an event takes place directly inside the monitored folder
 isDirectlyInside :: FilePath -> Event -> Bool
 isDirectlyInside dirPath event = isRelevantFileEvent || isRelevantDirEvent
   where
-    isRelevantFileEvent = (not $ eventIsDirectory event) && (takeDirectory dirPath == (takeDirectory $ eventPath event))
-    isRelevantDirEvent = eventIsDirectory event && (takeDirectory dirPath == (takeDirectory $ takeDirectory $ eventPath event))
+    isRelevantFileEvent = (eventIsDirectory event == IsFile) && (takeDirectory dirPath == (takeDirectory $ eventPath event))
+    isRelevantDirEvent = (eventIsDirectory event == IsDirectory) && (takeDirectory dirPath == (takeDirectory $ takeDirectory $ eventPath event))
 
-handleEvents :: Bool -> ActionPredicate -> EventChannel -> FilePath -> DebouncePayload -> [Event] -> IO ()
-handleEvents isRecursive actPred chan dirPath dbp (event:events) = do
-  when (actPred event && (isRecursive || (isDirectlyInside dirPath event))) $ case dbp of
-      (Just (DebounceData epsilon ior)) -> do
-        lastEvent <- readIORef ior
-        when (not $ debounce epsilon lastEvent event) (writeChan chan event)
-        atomicModifyIORef ior (\_ -> (event, ()))
-      Nothing -> writeChan chan event
-  handleEvents isRecursive actPred chan dirPath dbp events
-handleEvents _ _ _ _ _ [] = return ()
-
-listenFn :: (ActionPredicate -> EventChannel -> FilePath -> DebouncePayload -> FSE.Event -> IO a)
+listenFn :: (ActionPredicate -> EventCallback -> FilePath -> FSE.Event -> IO a)
          -> WatchConfig
          -> OSXManager
          -> FilePath
          -> ActionPredicate
-         -> EventChannel
+         -> EventCallback
          -> IO StopListening
-listenFn handler conf (OSXManager mvarMap) path actPred chan = do
+listenFn handler conf (OSXManager mvarMap) path actPred callback = do
   path' <- canonicalizeDirPath path
-  dbp <- newDebouncePayload $ confDebounce conf
   unique <- newUnique
-  eventStream <- FSE.eventStreamCreate [path'] 0.0 True False True (handler actPred chan path' dbp)
-  modifyMVar_ mvarMap $ \watchMap -> return (Map.insert unique (WatchData eventStream chan) watchMap)
+  eventStream <- FSE.eventStreamCreate [path'] 0.0 True False True (handler actPred callback path')
+  modifyMVar_ mvarMap $ \watchMap -> return (Map.insert unique (WatchData eventStream callback) watchMap)
   return $ do
     FSE.eventStreamDestroy eventStream
     modifyMVar_ mvarMap $ \watchMap -> return $ Map.delete unique watchMap
 
-instance FileListener OSXManager where
-  initSession = do
+instance FileListener OSXManager () where
+  initSession _ = do
     (v1, v2, _) <- FSE.osVersion
-    if not $ v1 > 10 || (v1 == 10 && v2 > 6) then return Nothing else
-      fmap (Just . OSXManager) $ newMVar Map.empty
+    if not $ v1 > 10 || (v1 == 10 && v2 > 6) then return $ Left "Unsupported OS version" else
+      (Right . OSXManager) <$> newMVar Map.empty
 
   killSession (OSXManager mvarMap) = do
     watchMap <- readMVar mvarMap
@@ -142,7 +134,5 @@ instance FileListener OSXManager where
       eventStreamDestroy' :: WatchData -> IO ()
       eventStreamDestroy' (WatchData eventStream _) = FSE.eventStreamDestroy eventStream
 
-  listen = listenFn $ handleEvent False
-  listenRecursive = listenFn $ handleEvent True
-
-  usesPolling = const False
+  listen = listenFn $ handleFSEEvent False
+  listenRecursive = listenFn $ handleFSEEvent True
