@@ -2,17 +2,17 @@
 -- Copyright (c) 2012 Mark Dittmer - http://www.markdittmer.org
 -- Developed for a Google Summer of Code project - http://gsoc2012.markdittmer.org
 --
+{-# LANGUAGE MultiParamTypeClasses #-}
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
-module System.FSNotify.Win32
-       ( FileListener(..)
-       , NativeManager
-       ) where
+module System.FSNotify.Win32 (
+  FileListener(..)
+  , NativeManager
+  ) where
 
 import Control.Concurrent
 import Control.Monad (when)
 import Data.Bits
-import Data.IORef (atomicModifyIORef, readIORef)
 import qualified Data.Map as Map
 import Data.Time (getCurrentTime, UTCTime)
 import Prelude
@@ -33,39 +33,28 @@ type BaseDir = FilePath
 -- handle[native]Event.
 
 -- Win32-notify has (temporarily?) dropped support for Renamed events.
-fsnEvent :: Bool -> BaseDir -> UTCTime -> WNo.Event -> Event
+fsnEvent :: EventIsDirectory -> BaseDir -> UTCTime -> WNo.Event -> Event
 fsnEvent isDirectory basedir timestamp (WNo.Created name) = Added (normalise (basedir </> name)) timestamp isDirectory
 fsnEvent isDirectory basedir timestamp (WNo.Modified name) = Modified (normalise (basedir </> name)) timestamp isDirectory
 fsnEvent isDirectory basedir timestamp (WNo.Deleted name) = Removed (normalise (basedir </> name)) timestamp isDirectory
 
-handleWNoEvent :: Bool -> BaseDir -> ActionPredicate -> EventChannel -> DebouncePayload -> WNo.Event -> IO ()
-handleWNoEvent isDirectory basedir actPred chan dbp inoEvent = do
+handleWNoEvent :: EventIsDirectory -> BaseDir -> ActionPredicate -> EventCallback -> WNo.Event -> IO ()
+handleWNoEvent isDirectory basedir actPred callback inoEvent = do
   currentTime <- getCurrentTime
   let event = fsnEvent isDirectory basedir currentTime inoEvent
-  handleEvent actPred chan dbp event
+  when (actPred event) $ callback event
 
-handleEvent :: ActionPredicate -> EventChannel -> DebouncePayload -> Event -> IO ()
-handleEvent actPred chan dbp event | actPred event = do
-  case dbp of
-    (Just (DebounceData epsilon ior)) -> do
-      lastEvent <- readIORef ior
-      when (not $ debounce epsilon lastEvent event) $ writeChan chan event
-      atomicModifyIORef ior (\_ -> (event, ()))
-    Nothing -> writeChan chan event
-handleEvent _ _ _ _ = return ()
-
-watchDirectory :: Bool -> WatchConfig -> WNo.WatchManager -> FilePath -> ActionPredicate -> EventChannel -> IO (IO ())
-watchDirectory isRecursive conf watchManager@(WNo.WatchManager mvarMap) path actPred chan = do
+watchDirectory :: Bool -> WatchConfig -> WNo.WatchManager -> FilePath -> ActionPredicate -> EventCallback -> IO (IO ())
+watchDirectory isRecursive conf watchManager@(WNo.WatchManager mvarMap) path actPred callback = do
   path' <- canonicalizeDirPath path
-  dbp <- newDebouncePayload $ confDebounce conf
 
   let fileFlags = foldl (.|.) 0 [WNo.fILE_NOTIFY_CHANGE_FILE_NAME, WNo.fILE_NOTIFY_CHANGE_SIZE, WNo.fILE_NOTIFY_CHANGE_ATTRIBUTES]
   let dirFlags = foldl (.|.) 0 [WNo.fILE_NOTIFY_CHANGE_DIR_NAME]
 
   -- Start one watch for file events and one for directory events
   -- (There seems to be no other way to provide isDirectory information)
-  wid1 <- WNo.watchDirectory watchManager path' isRecursive fileFlags (handleWNoEvent False path' actPred chan dbp)
-  wid2 <- WNo.watchDirectory watchManager path' isRecursive dirFlags (handleWNoEvent True path' actPred chan dbp)
+  wid1 <- WNo.watchDirectory watchManager path' isRecursive fileFlags (handleWNoEvent IsFile path' actPred callback)
+  wid2 <- WNo.watchDirectory watchManager path' isRecursive dirFlags (handleWNoEvent IsDirectory path' actPred callback)
 
   -- The StopListening action should make sure to remove the watches from the manager after they're killed.
   -- Otherwise, a call to killSession would cause us to try to kill them again, resulting in an invalid handle error.
@@ -76,15 +65,13 @@ watchDirectory isRecursive conf watchManager@(WNo.WatchManager mvarMap) path act
     WNo.killWatch wid2
     modifyMVar_ mvarMap $ \watchMap -> return (Map.delete wid2 watchMap)
 
-instance FileListener WNo.WatchManager where
+instance FileListener WNo.WatchManager () where
   -- TODO: This should actually lookup a Windows API version and possibly return
   -- Nothing the calls we need are not available. This will require that API
   -- version information be exposed by Win32-notify.
-  initSession = fmap Just WNo.initWatchManager
+  initSession _ = Right <$> WNo.initWatchManager
 
   killSession = WNo.killWatchManager
 
   listen = watchDirectory False
   listenRecursive = watchDirectory True
-
-  usesPolling = const False
